﻿////////////////////////////////////////////////////////////////////////////////////////////////////
// test_fastest_sort_wn.cpp
// Copyright(c) 2018 ozk
// This software is released under the MIT License.
// http://opensource.org/licenses/mit-license.php
////////////////////////////////////////////////////////////////////////////////////////////////////
#include <algorithm>
#include <array>
#include <cassert>
#include <iostream>
#include <string>
#include <vector>


namespace
{

	// 生成木のノード
	template<size_t N>
	struct Node
	{
		std::vector<std::array<int, N>> permutation; // 順列
		size_t lhs; // compare()の左手側の引数のインデックス
		size_t rhs; // compare()の右手側の引数のインデックス
		Node *child_less_equal; // perm[lhs] <= perm[rhs]を満たす子ノード
		Node *child_greater; // perm[lhs] > perm[rhs]を満たす子ノード

		// コンストラクタ
		Node()
			: permutation{}, lhs{ 0 }, rhs{ 0 }, child_less_equal{ nullptr }, child_greater{ nullptr }
		{}
	};

	// 最速ソートwn生成器
	template<size_t N>
	class Generator
	{
		Node<N> *root_; // 生成木のルート

	public:

		// コンストラクタ
		Generator()
			: root_{ nullptr }
		{}

		// デストラクタ
		~Generator()
		{
			deallocate_node(root_);
		}

		// 生成木を作る
		void generate_tree()
		{
			root_ = new Node<N>{};
			// すべての順列を生成する
			std::cout << "BEGIN generate_permutation" << std::endl;
			generate_permutation(root_->permutation);
			std::cout << "END generate_permutation" << std::endl;
			// 順列を絞り込む最適条件を探しながら、子ノードを作る
			std::cout << "BEGIN generate_node" << std::endl;
			generate_child_node(root_);
			std::cout << "END generate_node" << std::endl;
		}

		// ソースコードを印字する
		void print_source_code(std::ostream &os)
		{
			std::cout << "BEGIN print_node" << std::endl;
			print_node(os, root_, 0);
			std::cout << "END print_node" << std::endl;
		}

	private:

		// ノードを解放する
		static void deallocate_node(Node<N> *node)
		{
			if (node != nullptr) {
				deallocate_node(node->child_less_equal);
				deallocate_node(node->child_greater);
				delete node;
			}
		}

		// 全ての順列を生成する
		static void generate_permutation(std::vector<std::array<int, N>> &permutation)
		{
			std::array<int, N> a;
			for (size_t i = 0; i < N; ++i) {
				a[i] = int(i);
			}
			do {
				permutation.push_back(a);
			} while (std::next_permutation(a.begin(), a.end()));
		}

		// 子ノードを作る
		static void generate_child_node(Node<N> *node)
		{
			assert(node != nullptr);

			// 順列を絞り込む最適条件を探す
			find_best_comparision(node->permutation, node->lhs, node->rhs);

			// 見つけた条件で順列を分配する
			node->child_less_equal = new Node<N>{};
			node->child_greater = new Node<N>{};
			for (auto i_perm = node->permutation.begin(); i_perm != node->permutation.end(); ++i_perm) {
				if (i_perm->at(node->lhs) <= i_perm->at(node->rhs)) {
					// perm[lhs] <= perm[rhs]
					node->child_less_equal->permutation.push_back(*i_perm);
				}
				else {
					// perm[lhs] > perm[rhs]
					node->child_greater->permutation.push_back(*i_perm);
				}
			}

			// 1になるまで再帰で子ノードを作り続ける
			if (node->child_less_equal->permutation.size() > 1) {
				generate_child_node(node->child_less_equal);
			}
			if (node->child_greater->permutation.size() > 1) {
				generate_child_node(node->child_greater);
			}
		}

		// ノードに対応するソースコードを印字する
		static void print_node(std::ostream &os, Node<N> *node, int depth)
		{
			if (node != nullptr) {
				// 再帰呼び出しの深さ分のタブでインデントする
				const std::string indent(depth, '\t');
				if (node->child_greater != nullptr && node->child_less_equal != nullptr) {
					// 両方の子ノードがあるときはif-else文を印字する
					os << indent << "if (compare(v" << node->rhs << ", v" << node->lhs << ")) {" << std::endl;
					print_node(os, node->child_greater, depth + 1);
					os << indent << "} else { " << std::endl;
					print_node(os, node->child_less_equal, depth + 1);
					os << indent << "}" << std::endl;
				}
				else {
					if (node->child_greater != nullptr) {
						// 子ノードが片方だけのケースはあり得ない。念のためエラーメッセージを印字する
						os << "ERROR node->child_greater != nullptr" << std::endl;
					}
					if (node->child_less_equal != nullptr) {
						// 子ノードが片方だけのケースはあり得ない。念のためエラーメッセージを印字する
						os << "ERROR node->child_less_equal != nullptr" << std::endl;
					}
					if (node->permutation.size() != 1) {
						// サイズが1以外はあり得ない。念のためエラーメッセージを印字する
						os << "ERROR node->permutation.size() != 1" << std::endl;
					}
					else {
						// ソート順に並べるためのswap()を印字する
						print_swap(os, node->permutation.front(), indent);
					}
				}
			}
		}

		// ソート順に並べるためのswap()を印字する
		// 例えば、
		//   int v0 = 0, v1 = 2, v2 = 3, v3 = 1;
		//   sort(v0, v1, v2, v3);
		// のような呼び出しのとき、引数は、
		//   permutation[] = 0231
		// で、作業表の初期値は、
		//   variable_order[] = 0123 ※入力時の変数との対応 { v0 = v0, v1 = v1, v2 = v2, v3 = v3 }
		//   sort_order[]     = 0312 ※ソート結果 { v0, v3, v1, v2 }
		// にする。
		// 最小回数のswap()を印字して、
		//   variable_order[] = 0312 ※入力時の変数との対応 { v0 = v0, v1 = v3, v2 = v1, v3 = v2 }
		// になるようにすれば完了。
		static void print_swap(std::ostream &os, const std::array<int, N> &permutation, const std::string &indent)
		{
			// 作業表を初期化する
			std::array<size_t, N> sort_order;
			std::array<size_t, N> variable_order;
			for (size_t i = 0; i < N; ++i) {
				sort_order[permutation[i]] = i;
				variable_order[i] = i;
			}
			// swap()を印字する
			std::string sep = indent;
			for (size_t i = 0; i < N - 1; ++i) {
				if (variable_order[i] != sort_order[i]) {
					size_t k;
					for (k = i + 1; k < N; ++k) {
						if (variable_order[k] == sort_order[i]) {
							break;
						}
					}
					std::cout << sep << "swap(v" << i << ", v" << k << ");";
					sep = " ";
					std::swap(variable_order[i], variable_order[k]);
				}
			}
			// コメントを印字する（abcd形式）
			std::array<char, N> abc;
			for (size_t i = 0; i < N; ++i) {
				abc[permutation[i]] = char('a' + i);
			}
			os << sep << "// ";
			for (size_t i = 0; i < N; ++i) {
				os << abc[i];
			}
			os << std::endl;
		}

		// 順列を絞り込む最適条件を探す
		static void find_best_comparision(const std::vector<std::array<int, N>> &permutation, size_t &best_lhs, size_t &best_rhs)
		{
			size_t best_diff = std::numeric_limits<size_t>::max();
			// lhsとrhsの組み合わせについて繰り返す
			for (size_t lhs = 0; lhs < N - 1; ++lhs) {
				for (size_t rhs = lhs + 1; rhs < N; ++rhs) {
					// perm[lhs] <= perm[rhs]を満たす順列を数える
					size_t count = 0;
					for (auto i_perm = permutation.begin(); i_perm != permutation.end(); ++i_perm) {
						if (i_perm->at(lhs) <= i_perm->at(rhs)) {
							++count;
						}
					}
					// 2*countがsizeになるのが最も良いので、2*countとsizeの差を求める
					size_t size = permutation.size();
					size_t diff = ((size > (2 * count)) ? (size - (2 * count)) : ((2 * count) - size));
					if (diff < best_diff) {
						// より良い条件が見つかった
						best_diff = diff;
						best_lhs = lhs;
						best_rhs = rhs;
					}
					if (best_diff == (permutation.size() % 2)) {
						// size % 2より良くなることはないので探索を打ち切る
						return;
					}
				}
			}
		}
	};

} // unnamed


int main()
{
	const size_t N = 4;
	Generator<N> gen;

	// 生成木を作る
	gen.generate_tree();

	// ソースコードを印字する
	gen.print_source_code(std::cout);

	return 0;
}
